/**
 * Quad Sim v0.21
 * Copyright (C) 2005  Giliam de Carpentier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * You can contact me at giliam@decarpentier.nl for questions concerning this license.
 */
#include "common.h"
#include <string.h>


// defs
#define MAPSIZE			256							// tiles
#define MAPSIZEPOWER	8							// 2log(MAPSIZE)
#define TILESIZE		.4f
#define MAPHEIGHT		75.f
#define RENDERTILES		30

// macros
#define MAPW(u, v)		pfMap[(((u) & (MAPSIZE - 1)) << MAPSIZEPOWER) + ((v) & (MAPSIZE - 1))]
#define MAPWN1(u, v)	pvMapN1[(((u) & (MAPSIZE - 1)) << MAPSIZEPOWER) + ((v) & (MAPSIZE - 1))]
#define MAPWN2(u, v)	pvMapN2[(((u) & (MAPSIZE - 1)) << MAPSIZEPOWER) + ((v) & (MAPSIZE - 1))]

V vGripTot;

V MAP::m2w(int u, int v) {
	return V((u + 0.5f * v) * TILESIZE, MAPW(u, v), -HALFSQRT3 * v * TILESIZE);
}

VMAP MAP::w2m(const V &v) {
	return VMAP(FLOAT2INT((v.x + THIRDSQRT3 * v.z) / TILESIZE),
				FLOAT2INT(-TWOTHIRDSSQRT3 * v.z / TILESIZE));
}

// return value:    0 --> Succesful
//					1 --> Error opening file
//					2 --> Invalid file
int MAP::readFile (const char *pcFN) {
	FILE *pFile;
    char pcBuffer[8];

	pFile = fopen (pcFN,"rb");
	if (pFile == NULL) return 1; // error opering file
    if (fread(&pcBuffer, 1, 8, pFile) != 8) return 2; // impossible size
    if (*(int *)&pcBuffer[4] != MAPSIZE) return 3; // wrong mapsize

	pcBuffer[4] = 0;
	if (strcmp(pcBuffer, "CPHM") != 0) return 4; // bad header

    for (int i = 0; i < MAPSIZE * MAPSIZE; i++) {
    	if (fread(pcBuffer, 1, 2, pFile) != 2) return 5; // file too small
        pfMap[i] = ((float)*(signed short int*)pcBuffer) / 65536.f * MAPHEIGHT * TILESIZE;
    }

	fclose(pFile);	
	return 0; // succesful
}


MAP::MAP(const char *pcFN) {
    pfMap = new float [MAPSIZE * MAPSIZE];
	printf ("reading map file ");
	printf ("[%d]\n",readFile(pcFN));

	/////////////////
	for (int j = 0; j < MAPSIZE; j++)
		for (int i = 0; i < MAPSIZE; i++) {
//			float h = 4 * sqrt(.5f - powf((float)i / MAPSIZE - .5f, 2) -
//					          powf((float)j / MAPSIZE - .5f, 2));
//			MAPW(i,j) -= 2 * h;
			MAPW(i,j) *= fabs((float)j/MAPSIZE-.5f)*1.8f+.1f;
		}
	/////////////////

	printf ("setting up map's normal tables ");
    pvMapN1 = new V [MAPSIZE * MAPSIZE];
    pvMapN2 = new V [MAPSIZE * MAPSIZE];
	
	V v = V(0.0f, HALFSQRT3 * TILESIZE, 0.0f);
	VMAP vm;

	for (vm.v = 0; vm.v < MAPSIZE; vm.v++)
		for (vm.u = 0; vm.u < MAPSIZE; vm.u++) {
			float h1 = MAPW(vm.u, vm.v);
			float h2 = MAPW(vm.u + 1, vm.v);
			float h3 = MAPW(vm.u, vm.v + 1);
			float h4 = MAPW(vm.u + 1, vm.v + 1);
			// left triangle
			v.x = HALFSQRT3 * (h1 - h2);
			v.z = h3 - 0.5f * (h1 + h2);
			MAPWN1(vm.u, vm.v) = v.normal();
			// right triangle
			v.x = HALFSQRT3 * (h3 - h4);
			v.z = 0.5f * (h3 + h4) - h2;
			MAPWN2(vm.u, vm.v) = v.normal();
		}
	printf ("[done]\n");
}

float MAP::heightAt(float x, float z) {
	float xt = x / TILESIZE;
	float zt = -z / TILESIZE;
	VMAP vm = VMAP(FLOAT2INT(xt - THIRDSQRT3 * zt),
			       FLOAT2INT(TWOTHIRDSSQRT3 * zt));
	float xta = vm.u + 0.5f * vm.v;
	float zta = HALFSQRT3 * vm.v;

	float xtr, ztr;
	float h1, h2, h3;
	
	if ((xt - xta) + THIRDSQRT3 * (zt - zta) < 1.0f) {
		// left triangle
		xtr = xta + 0.5f - xt;
		ztr = zta + HALFSQRT3 - zt;
		h1 = MAPW(vm.u, vm.v + 1);
		h2 = MAPW(vm.u, vm.v);
		h3 = MAPW(vm.u + 1, vm.v);
	} else {
		// right triangle
		xtr = xt - xta - 1.0f;
		ztr = zt - zta;
		h1 = MAPW(vm.u + 1, vm.v);
		h2 = MAPW(vm.u + 1, vm.v + 1);
		h3 = MAPW(vm.u, vm.v + 1);
	}

	return h1 + (h2 - h3) * xtr + THIRDSQRT3 * (h2 + h3 - 2 * h1) * ztr;
}

float MAP::tileDist(const V &v, V *pvRet) {
	VMAP vmap = w2m(v);
	V vD = v - m2w(vmap.u, vmap.v), vN;

	if (vD.x - THIRDSQRT3 * vD.z < TILESIZE) {
		// left triangle
		vN = MAPWN1(vmap.u, vmap.v);
	} else {
		// right triangle
		vD.x -= TILESIZE;
		vD.y = v.y - MAPW(vmap.u + 1, vmap.v);
		vN = MAPWN2(vmap.u, vmap.v);
	}
	if (pvRet != NULL)
		*pvRet = vN;
	return vD & vN;
}


V MAP::force(const MP *pmp) {
	V vNorm;
	float dist = tileDist(pmp->vP, &vNorm);
    if (dist < .04f) {
		float normV = pmp->vV & vNorm, normF = pmp->vF & vNorm, normTileF = 0;
		V vPlaneV = pmp->vV - normV * vNorm, vPlaneF = pmp->vF - normF * vNorm;
		V vFricF = -pmp->m / TAU * vPlaneV - vPlaneF;
		if (dist < -.02f) normTileF -= 50.f * pmp->m * (dist + .02f);
		if (normF < .0f) normTileF -= normF;
		if (normV < .0f) normTileF -= pmp->m / TAU * normV;
		float fricF = vFricF.len();
		float fricFMax = normTileF * (0.45f + .007f / (.01f + (vPlaneV & vPlaneV)));
		if (fricF > fricFMax) vFricF *= fricFMax / fricF;
		return (dist < .0f ? normTileF * vNorm : VZERO) + vFricF;
	}
	return VZERO;
}

V MAP::force(const MP *pmp, float cFric, float cFric0, V vProf, float cFricProf) {
	V vNorm;
	float dist = tileDist(pmp->vP, &vNorm);
    if (dist < .04f) {
		float normV = pmp->vV & vNorm, normF = pmp->vF & vNorm, normTileF = 0;
		V vPlaneV = pmp->vV - normV * vNorm, vPlaneF = pmp->vF - normF * vNorm;
		V vFricF = -pmp->m / TAU * vPlaneV - vPlaneF;
		if (dist < -.02f) normTileF -= 50.f * pmp->m * (dist + .02f);
		if (normF < .0f) normTileF -= normF;
		if (normV < .0f) normTileF -= pmp->m / TAU * normV;
		float fricF = vFricF.len();
		float fricFMax = normTileF * (cFric + cFric0 / (1.f + 100.f * (vPlaneV & vPlaneV)));
		if (fricF > fricFMax) {
			vGripTot += vFricF;
			V vDiffF = vFricF - vFricF * fricFMax / fricF; vFricF -= vDiffF;
			vProf = (vProf - vNorm * (vProf & vNorm)).normal();
			float profF =  vProf & vDiffF, profFMax = normTileF * cFricProf;
			if (profF >= 0.f) vFricF += vProf * ((profF > profFMax) ? profFMax : profF);
			else			  vFricF += vProf * ((-profF > profFMax) ? -profFMax : profF);
			vGripTot -= vFricF;
		}
		return (dist < .0f ? normTileF * vNorm : VZERO) + vFricF;
	}
	return VZERO;
}


#define	RENDERRADIUS		(HALFSQRT3 * TILESIZE * (RENDERTILES-2))
#define RENDERRADIUS2		(RENDERRADIUS * RENDERRADIUS)

V MAP::vertexColor(const V &v, bool offset) {
	float d2, p;
	d2 = v.x * v.x + v.z * v.z;
	p = MIN(d2 / RENDERRADIUS2, 1.0f);
//	V vret = offset ? V(0.97f, 0.94f, 0.35f) : V(1.f, 1.f, 0.5f); // sand
//	V vret = offset ? V(.5f, 0.3f, 0.0f) : V(0.6f, 0.4f, 0.1f); // mud
//	V vret = offset ? V(.5f, 0.65f, 0.15f) : V(0.58f, 0.54f, 0.12f); // patch
//  V vret = offset ? V(.65f, 0.45f, 0.25f) : V(0.75f, 0.55f, 0.35f);

	V vret = offset ? V(.5f, 0.4f, 0.25f) : V(0.35f, 0.35f, 0.3f);
	return vret * (1-p) + V(0.90f, 0.95f, 0.98f) * p;
}


void MAP::render(const V &vc) {
	VMAP vmap = w2m(vc);
	V vxz = vc * V(1.0f, 0.0f, 1.0f);

	for (int rv = 0; rv < RENDERTILES; rv++) {
		V v; int ru;
		glBegin(GL_TRIANGLE_STRIP);
			v = m2w(vmap.u + RENDERTILES - rv, rv + vmap.v);
			glColor3fv(vertexColor(v - vxz, false).p);
			glVertex3fv(v.p);
			for (ru = RENDERTILES - rv - 1; ru >= -RENDERTILES; ru--) {
				v = m2w(ru + vmap.u, rv + vmap.v + 1);
				glColor3fv(vertexColor(v - vxz, false).p);
				glVertex3fv(v.p);
				v = m2w(ru + vmap.u, rv + vmap.v);
				glColor3fv(vertexColor(v - vxz, true).p);
				glVertex3fv(v.p);
			}
		glEnd();
		glBegin(GL_TRIANGLE_STRIP);
			v = m2w(rv + vmap.u - RENDERTILES, vmap.v - rv);
			glColor3fv(vertexColor(v - vxz, false).p);
			glVertex3fv(v.p);
			for (ru = rv - RENDERTILES; ru < RENDERTILES; ru++) {
				v = m2w(ru + vmap.u + 1, -rv + vmap.v - 1);
				glColor3fv(vertexColor(v - vxz, true).p);
				glVertex3fv(v.p);
				v = m2w(ru + vmap.u + 1, -rv + vmap.v);
				glColor3fv(vertexColor(v - vxz, false).p);
				glVertex3fv(v.p);
			}
		glEnd();
	}
}














