/**
 * Quad Sim v0.21
 * Copyright (C) 2005  Giliam de Carpentier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * You can contact me at giliam@decarpentier.nl for questions concerning this license.
 */
// included by "common.h"
#ifndef MATHS_H
#define MATHS_H

// ----------------------------------------------------------------------------
// CONTSTANTS
// ----------------------------------------------------------------------------
#define HALFSQRT3		0.866025403f
#define THIRDSQRT3		0.577350269f
#define TWOTHIRDSSQRT3	1.154700538f
#define TWOPI			6.283185307f
#define PI				3.141592653f
#define HALFPI			1.570796327f

// ----------------------------------------------------------------------------
// MACROS
// ----------------------------------------------------------------------------
#define MIN(a, b)		((a)<(b)?a:b)
#define MAX(a, b)		((a)>(b)?a:b)
#define ABS(a)			((a)>=.0f?a:-(a))
#define FLOAT2INT(a)	((a)>.0f?(int)(a):(int)(a)-1)
#define FILTER(a,b,c)   a=a*(c)+(b)*(1.f-(c))

// ----------------------------------------------------------------------------
// CLASS: V (Vector) declaration & implementation
// DESCR: Complete 3D vector implementation with lots of operators and methods
// ----------------------------------------------------------------------------
#define VZERO ((const V)V(0.0f, 0.0f, 0.0f))

class V {
public:
	union {
		struct {
			float p[3];
		};
		struct {
			float x, y, z;
		};
	};

	V() {}
	
	V(float x, float y, float z) {
		this->x = x; this->y = y; this->z = z;}

	static friend V operator-(const V &v) {
		return V(-v.x, -v.y, -v.z);}

	static friend V operator+(const V &a, const V &b) {
	    return V(a.x + b.x, a.y + b.y, a.z + b.z);}

	static friend V operator-(const V &a, const V &b) {
	    return V(a.x - b.x, a.y - b.y, a.z - b.z);}

	static friend V operator*(float d, const V &a) {
	    return V(d * a.x, d * a.y, d * a.z);}

	static friend V operator*(const V &a, float d) {
	    return V(a.x * d, a.y * d, a.z * d);}

	static friend V operator*(const V &a, const V &b) {
	    return V(a.x * b.x, a.y * b.y, a.z * b.z);}

	static friend V operator/(float d, const V &a) {
		return V(d / a.x, d / a.y, d / a.z);}

	static friend V operator/(const V &a, float d) {
		return V(a.x / d, a.y / d, a.z / d);}

	static friend V operator/(const V &a, const V &b) {
	    return V(a.x / b.x, a.y / b.y, a.z / b.z);}

	static friend float operator&(const V &a, const V &b) {
	    return a.x * b.x + a.y * b.y + a.z * b.z;}

	static friend V operator^(const V &a, const V &b) {
	    return V(a.y * b.z - a.z * b.y,
		         a.z * b.x - a.x * b.z,
		         a.x * b.y - a.y * b.x);}

	static friend V operator+=(V &a, const V &b)	{return a = a + b;}
	static friend V operator-=(V &a, const V &b)	{return a = a - b;}
	static friend V operator*=(V &a, float d)		{return a = a * d;}
	static friend V operator*=(V &a, const V &b)	{return a = a * b;}
	static friend V operator/=(V &a, float d)		{return a = a / d;}
	static friend V operator/=(V &a, const V &b)	{return a = a / b;}
	static friend V operator^=(V &a, const V &b)	{return a = a ^ b;}

	float len() {return sqrtf(x * x + y * y + z * z);}
	V normal() {return V(x,y,z) / len();}
	float smallest() {return (x < y ? MIN(x, y) : MIN(y, z));}
	float largest() {return (x < y ? MAX(y, z) : MAX(x, z));}

	void dump() {
		printf ("<V: %f %f %f>\n", x, y, z);}
};


// ----------------------------------------------------------------------------
// CLASS: RM (Rotation Matrix) declaration
// DESCR: Can define, invert, multiply and map 3D rotations
// ----------------------------------------------------------------------------
class RM {
public:
	V vA, vB, vC;
	RM(const V &vA, const V &vB, const V &vC) {
		this->vA = vA; this->vB = vB, this->vC = vC;
	}

	RM() {}

	RM(const V &v) {
		V vu = v; float r = vu.len();
		if (r == 0.f) {
			vA = V(1.f, 0.f, 0.f);
			vB = V(0.f, 1.f, 0.f);
			vC = V(0.f, 0.f, 1.f);
		} else {
			vu /= r; float s = sinf(r), c = cosf(r);
			float xcc = vu.x * (1-c), xs = vu.x * s;
			float ycc = vu.y * (1-c), ys = vu.y * s;
			float zcc = vu.z * (1-c), zs = vu.z * s;
			vA = V(xcc * vu.x + c, xcc * vu.y + zs, xcc * vu.z - ys);
			vB = V(ycc * vu.x - zs, ycc * vu.y + c, ycc * vu.z + xs);
			vC = V(zcc * vu.x + ys, zcc * vu.y - xs, zcc * vu.z + c);
		}
	}

	static friend RM operator!(const RM &rm) {
		return RM(V(rm.vA.x, rm.vB.x, rm.vC.x), 
				  V(rm.vA.y, rm.vB.y, rm.vC.y),
				  V(rm.vA.z, rm.vB.z, rm.vC.z));
	}

	static friend RM operator*(const RM &a, const RM &b) {
		RM rm = RM(V(a.vA.x * b.vA.x + a.vB.x * b.vA.y + a.vC.x * b.vA.z,
					 a.vA.y * b.vA.x + a.vB.y * b.vA.y + a.vC.y * b.vA.z,
					 a.vA.z * b.vA.x + a.vB.z * b.vA.y + a.vC.z * b.vA.z),
				   V(a.vA.x * b.vB.x + a.vB.x * b.vB.y + a.vC.x * b.vB.z,
					 a.vA.y * b.vB.x + a.vB.y * b.vB.y + a.vC.y * b.vB.z,
					 a.vA.z * b.vB.x + a.vB.z * b.vB.y + a.vC.z * b.vB.z),
				   V(a.vA.x * b.vC.x + a.vB.x * b.vC.y + a.vC.x * b.vC.z,
					 a.vA.y * b.vC.x + a.vB.y * b.vC.y + a.vC.y * b.vC.z,
					 a.vA.z * b.vC.x + a.vB.z * b.vC.y + a.vC.z * b.vC.z));
		rm.vA = rm.vA.normal();
		rm.vC = (rm.vA ^ rm.vB).normal();
		rm.vB = (rm.vC ^ rm.vA).normal();
		return rm;
	}

	static friend RM operator*=(RM &a, const RM &b) {return a = a * b;}

		static friend V operator*(const RM &rm, const V &v) {
	    return V(rm.vA.x * v.x + rm.vB.x * v.y + rm.vC.x * v.z,
				 rm.vA.y * v.x + rm.vB.y * v.y + rm.vC.y * v.z,
				 rm.vA.z * v.x + rm.vB.z * v.y + rm.vC.z * v.z);
	}


	V vector() {
		V v = V(vB.z - vC.y, vC.x - vA.z, vA.y - vB.x);
		float r = v.len();
		if (r == 0.0f) return VZERO;
		if (vA.x + vB.y + vC.z < 1.0f)
			return v * (PI - asinf(r / 2)) / r;
		return v * asinf(r / 2) / r;
	}

	void dump() {
		printf ("<RM:(%f %f %f)\n", vA.x, vB.x, vC.x);
		printf ("    (%f %f %f)\n", vA.y, vB.y, vC.y);
		printf ("    (%f %f %f)>\n",vA.z, vB.z, vC.z);
	}

};

#endif