/**
 * Quad Sim v0.21
 * Copyright (C) 2005  Giliam de Carpentier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * You can contact me at giliam@decarpentier.nl for questions concerning this license.
 */
// included by "common.h"
#ifndef PHYSICS_H
#define PHYSICS_H

// ----------------------------------------------------------------------------
// CONTSTANTS
// ----------------------------------------------------------------------------
#define TAU				0.001f

// ----------------------------------------------------------------------------
// CLASS: PVV (Position & Velocity Vector) declaration & implementation
// DESCR: Superclass of MP and BODY
// ----------------------------------------------------------------------------
class PVV {
public:
	V vP, vV;

	PVV() {}
	PVV(const V &vP, const V &vV) {
		this->vP = vP; this->vV = vV;}
};

// ----------------------------------------------------------------------------
// CLASS: MP (Mass Point)
// DESCR: Can define, advance in time and calculate common forces for a mp
// ----------------------------------------------------------------------------
class MP : public PVV {
protected:
	MP(float m) {this->m = m; vF = VZERO;}
public:
	V vF;
	float m;

	MP(const PVV *ppvv, float m) : PVV(ppvv->vP, ppvv->vV) {
		this->m = m; vF = VZERO;}

	void addForce(const V &vF) {
		this->vF += vF;}

	void advance() {
		vP += (vV += vF / m * TAU) * TAU; vF = VZERO;}
};

// ----------------------------------------------------------------------------
// CLASS: MB (Mass Body)
// DESCR: simulates a spring between two mass points
// ----------------------------------------------------------------------------
class MB : public MP {
protected:
	MB(float m, const V &vTheta);
public:
	V vM, vOmega, vTheta;
	RM rmPhi, rmPhiT, rmDelta; 

	MB(const PVV &pvv, const PVV &pvvRot, float m, const V &vTheta);
	V w2m(const V &v);
	V m2w(const V &v);
	PVV m2pvv(const V &v);
	void addForce(const V &vF);
	void addMoment(const V &vM);
	void addForce(const V &vP, const V &vF);
	void advance();
};


// ----------------------------------------------------------------------------
// CLASS: GRAVITY
// DESCR: returns force of gravity for mass points
// ----------------------------------------------------------------------------
class GRAVITY {
	float c;
public:
	GRAVITY(float constant) {
		c = -constant;
	}

	V force(const MP *pmp) {
		return V(0.0f, c * pmp->m, 0.0f);
	}
};


// ----------------------------------------------------------------------------
// CLASS: SPRINGPP
// DESCR: simulates a spring between two mass points
// ----------------------------------------------------------------------------
class SPRINGPP {
protected:
	void init(float max, float fS, float fD);
	SPRINGPP() {}
public:
	float length, cS, cD;
	PVV *ppvvA, *ppvvB;
	SPRINGPP(MP *pmpA, MP *pmpB, float fS, float fD);
	V force();
	void addForce(const V &vF);
};


// ----------------------------------------------------------------------------
// CLASS: SPRINGBP
// DESCR: simulates a spring between a mass body and a mass point
// ----------------------------------------------------------------------------
class SPRINGBP : public SPRINGPP {
public:
	MB *pmbA;
	V vA;

	SPRINGBP(MB *pmbA, const V &vA, MP *pmpB, float fS, float fD);
	~SPRINGBP();
	V force();
	void addForce(const V &vF);
};

// ----------------------------------------------------------------------------
// CLASS: SPRINGBB
// DESCR: simulates a spring between a two mass bodies
// ----------------------------------------------------------------------------
class SPRINGBB : public SPRINGPP {
public:
	MB *pmbA, *pmbB;
	V vA, vB;

	SPRINGBB(MB *pmbA, const V &vA, MB *pmbB, const V &vB, float fS, float fD);
	~SPRINGBB();
	V force();
	void addForce(const V &vF);
};



// ----------------------------------------------------------------------------
// CLASS: RMB (Relative Mass Body) declaration
// DESCR: defines a child MB relative to a parent MB
// ----------------------------------------------------------------------------
class RMB : public MB {
public:
	MB *pmb;
	V vRelP;

	RMB(MB *pmb, const V &vRelP);
	PVV m2pvv(const V &v);
	void pre();
	void advance();
	void post();
};

#endif
